@@ -21,25 +21,42 @@ The :mod:`gridfs` package is an implementation of GridFS on top of
 """
 
 from collections import abc
+from typing import Any, List, Mapping, Optional, cast
 
-from pymongo import (ASCENDING,
-                     DESCENDING)
-from pymongo.common import UNAUTHORIZED_CODES, validate_string
-from pymongo.database import Database
-from pymongo.errors import ConfigurationError, OperationFailure
-
+from bson.objectid import ObjectId
 from gridfs.errors import NoFile
-from gridfs.grid_file import (GridIn,
-                              GridOut,
-                              GridOutCursor,
-                              DEFAULT_CHUNK_SIZE,
-                              _clear_entity_type_registry,
-                              _disallow_transactions)
+from gridfs.grid_file import (
+    DEFAULT_CHUNK_SIZE,
+    GridIn,
+    GridOut,
+    GridOutCursor,
+    _clear_entity_type_registry,
+    _disallow_transactions,
+)
+from pymongo import ASCENDING, DESCENDING, _csot
+from pymongo.client_session import ClientSession
+from pymongo.collection import Collection
+from pymongo.common import validate_string
+from pymongo.database import Database
+from pymongo.errors import ConfigurationError
+from pymongo.read_preferences import _ServerMode
+from pymongo.write_concern import WriteConcern
+
+__all__ = [
+    "GridFS",
+    "GridFSBucket",
+    "NoFile",
+    "DEFAULT_CHUNK_SIZE",
+    "GridIn",
+    "GridOut",
+    "GridOutCursor",
+]
+
 
 class GridFS(object):
-    """An instance of GridFS on top of a single Database.
-    """
-    def __init__(self, database, collection="fs"):
+    """An instance of GridFS on top of a single Database."""
+
+    def __init__(self, database: Database, collection: str = "fs"):
         """Create a new instance of :class:`GridFS`.
 
         Raises :class:`TypeError` if `database` is not an instance of
@@ -75,14 +92,13 @@ class GridFS(object):
         database = _clear_entity_type_registry(database)
 
         if not database.write_concern.acknowledged:
-            raise ConfigurationError('database must use '
-                                     'acknowledged write_concern')
+            raise ConfigurationError("database must use acknowledged write_concern")
 
         self.__collection = database[collection]
         self.__files = self.__collection.files
         self.__chunks = self.__collection.chunks
 
-    def new_file(self, **kwargs):
+    def new_file(self, **kwargs: Any) -> GridIn:
         """Create a new file in GridFS.
 
         Returns a new :class:`~gridfs.grid_file.GridIn` instance to
@@ -98,16 +114,13 @@ class GridFS(object):
         """
         return GridIn(self.__collection, **kwargs)
 
-    def put(self, data, **kwargs):
+    def put(self, data: Any, **kwargs: Any) -> Any:
         """Put data in GridFS as a new file.
 
         Equivalent to doing::
 
-          try:
-              f = new_file(**kwargs)
+          with fs.new_file(**kwargs) as f:
               f.write(data)
-          finally:
-              f.close()
 
         `data` can be either an instance of :class:`bytes` or a file-like
         object providing a :meth:`read` method. If an `encoding` keyword
@@ -128,15 +141,12 @@ class GridFS(object):
         .. versionchanged:: 3.0
            w=0 writes to GridFS are now prohibited.
         """
-        grid_file = GridIn(self.__collection, **kwargs)
-        try:
+
+        with GridIn(self.__collection, **kwargs) as grid_file:
             grid_file.write(data)
-        finally:
-            grid_file.close()
+            return grid_file._id
 
-        return grid_file._id
-
-    def get(self, file_id, session=None):
+    def get(self, file_id: Any, session: Optional[ClientSession] = None) -> GridOut:
         """Get a file from GridFS by ``"_id"``.
 
         Returns an instance of :class:`~gridfs.grid_file.GridOut`,
@@ -156,7 +166,13 @@ class GridFS(object):
         gout._ensure_file()
         return gout
 
-    def get_version(self, filename=None, version=-1, session=None, **kwargs):
+    def get_version(
+        self,
+        filename: Optional[str] = None,
+        version: Optional[int] = -1,
+        session: Optional[ClientSession] = None,
+        **kwargs: Any
+    ) -> GridOut:
         """Get a file from GridFS by ``"filename"`` or metadata fields.
 
         Returns a version of the file in GridFS whose filename matches
@@ -197,6 +213,8 @@ class GridFS(object):
 
         _disallow_transactions(session)
         cursor = self.__files.find(query, session=session)
+        if version is None:
+            version = -1
         if version < 0:
             skip = abs(version) - 1
             cursor.limit(-1).skip(skip).sort("uploadDate", DESCENDING)
@@ -204,12 +222,13 @@ class GridFS(object):
             cursor.limit(-1).skip(version).sort("uploadDate", ASCENDING)
         try:
             doc = next(cursor)
-            return GridOut(
-                self.__collection, file_document=doc, session=session)
+            return GridOut(self.__collection, file_document=doc, session=session)
         except StopIteration:
             raise NoFile("no version %d for filename %r" % (version, filename))
 
-    def get_last_version(self, filename=None, session=None, **kwargs):
+    def get_last_version(
+        self, filename: Optional[str] = None, session: Optional[ClientSession] = None, **kwargs: Any
+    ) -> GridOut:
         """Get the most recent version of a file in GridFS by ``"filename"``
         or metadata fields.
 
@@ -228,7 +247,7 @@ class GridFS(object):
         return self.get_version(filename=filename, session=session, **kwargs)
 
     # TODO add optional safe mode for chunk removal?
-    def delete(self, file_id, session=None):
+    def delete(self, file_id: Any, session: Optional[ClientSession] = None) -> None:
         """Delete a file from GridFS by ``"_id"``.
 
         Deletes all data belonging to the file with ``"_id"``:
@@ -257,7 +276,7 @@ class GridFS(object):
         self.__files.delete_one({"_id": file_id}, session=session)
         self.__chunks.delete_many({"files_id": file_id}, session=session)
 
-    def list(self, session=None):
+    def list(self, session: Optional[ClientSession] = None) -> List[str]:
         """List the names of all files stored in this instance of
         :class:`GridFS`.
 
@@ -275,10 +294,16 @@ class GridFS(object):
         # With an index, distinct includes documents with no filename
         # as None.
         return [
-            name for name in self.__files.distinct("filename", session=session)
-            if name is not None]
+            name for name in self.__files.distinct("filename", session=session) if name is not None
+        ]
 
-    def find_one(self, filter=None, session=None, *args, **kwargs):
+    def find_one(
+        self,
+        filter: Optional[Any] = None,
+        session: Optional[ClientSession] = None,
+        *args: Any,
+        **kwargs: Any
+    ) -> Optional[GridOut]:
         """Get a single file from gridfs.
 
         All arguments to :meth:`find` are also valid arguments for
@@ -311,7 +336,7 @@ class GridFS(object):
 
         return None
 
-    def find(self, *args, **kwargs):
+    def find(self, *args: Any, **kwargs: Any) -> GridOutCursor:
         """Query GridFS for files.
 
         Returns a cursor that iterates across files matching
@@ -343,9 +368,9 @@ class GridFS(object):
         are associated with that session.
 
         :Parameters:
-          - `filter` (optional): a SON object specifying elements which
-            must be present for a document to be included in the
-            result set
+          - `filter` (optional): A query document that selects which files
+            to include in the result set. Can be an empty document to include
+            all files.
           - `skip` (optional): the number of files to omit (from
             the start of the result set) when returning the results
           - `limit` (optional): the maximum number of results to
@@ -372,7 +397,12 @@ class GridFS(object):
         """
         return GridOutCursor(self.__collection, *args, **kwargs)
 
-    def exists(self, document_or_id=None, session=None, **kwargs):
+    def exists(
+        self,
+        document_or_id: Optional[Any] = None,
+        session: Optional[ClientSession] = None,
+        **kwargs: Any
+    ) -> bool:
         """Check if a file exists in this instance of :class:`GridFS`.
 
         The file to check for can be specified by the value of its
@@ -422,9 +452,14 @@ class GridFS(object):
 class GridFSBucket(object):
     """An instance of GridFS on top of a single Database."""
 
-    def __init__(self, db, bucket_name="fs",
-                 chunk_size_bytes=DEFAULT_CHUNK_SIZE, write_concern=None,
-                 read_preference=None):
+    def __init__(
+        self,
+        db: Database,
+        bucket_name: str = "fs",
+        chunk_size_bytes: int = DEFAULT_CHUNK_SIZE,
+        write_concern: Optional[WriteConcern] = None,
+        read_preference: Optional[_ServerMode] = None,
+    ) -> None:
         """Create a new instance of :class:`GridFSBucket`.
 
         Raises :exc:`TypeError` if `database` is not an instance of
@@ -466,22 +501,28 @@ class GridFSBucket(object):
 
         wtc = write_concern if write_concern is not None else db.write_concern
         if not wtc.acknowledged:
-            raise ConfigurationError('write concern must be acknowledged')
+            raise ConfigurationError("write concern must be acknowledged")
 
         self._bucket_name = bucket_name
         self._collection = db[bucket_name]
-        self._chunks = self._collection.chunks.with_options(
-            write_concern=write_concern,
-            read_preference=read_preference)
+        self._chunks: Collection = self._collection.chunks.with_options(
+            write_concern=write_concern, read_preference=read_preference
+        )
 
-        self._files = self._collection.files.with_options(
-            write_concern=write_concern,
-            read_preference=read_preference)
+        self._files: Collection = self._collection.files.with_options(
+            write_concern=write_concern, read_preference=read_preference
+        )
 
         self._chunk_size_bytes = chunk_size_bytes
+        self._timeout = db.client.options.timeout
 
-    def open_upload_stream(self, filename, chunk_size_bytes=None,
-                           metadata=None, session=None):
+    def open_upload_stream(
+        self,
+        filename: str,
+        chunk_size_bytes: Optional[int] = None,
+        metadata: Optional[Mapping[str, Any]] = None,
+        session: Optional[ClientSession] = None,
+    ) -> GridIn:
         """Opens a Stream that the application can write the contents of the
         file to.
 
@@ -492,11 +533,11 @@ class GridFSBucket(object):
 
           my_db = MongoClient().test
           fs = GridFSBucket(my_db)
-          grid_in = fs.open_upload_stream(
+          with fs.open_upload_stream(
                 "test_file", chunk_size_bytes=4,
-                metadata={"contentType": "text/plain"})
-          grid_in.write("data I want to store!")
-          grid_in.close()  # uploaded on close
+                metadata={"contentType": "text/plain"}) as grid_in:
+              grid_in.write("data I want to store!")
+          # uploaded on close
 
         Returns an instance of :class:`~gridfs.grid_file.GridIn`.
 
@@ -519,17 +560,25 @@ class GridFSBucket(object):
         """
         validate_string("filename", filename)
 
-        opts = {"filename": filename,
-                "chunk_size": (chunk_size_bytes if chunk_size_bytes
-                               is not None else self._chunk_size_bytes)}
+        opts = {
+            "filename": filename,
+            "chunk_size": (
+                chunk_size_bytes if chunk_size_bytes is not None else self._chunk_size_bytes
+            ),
+        }
         if metadata is not None:
             opts["metadata"] = metadata
 
         return GridIn(self._collection, session=session, **opts)
 
     def open_upload_stream_with_id(
-            self, file_id, filename, chunk_size_bytes=None, metadata=None,
-            session=None):
+        self,
+        file_id: Any,
+        filename: str,
+        chunk_size_bytes: Optional[int] = None,
+        metadata: Optional[Mapping[str, Any]] = None,
+        session: Optional[ClientSession] = None,
+    ) -> GridIn:
         """Opens a Stream that the application can write the contents of the
         file to.
 
@@ -540,13 +589,13 @@ class GridFSBucket(object):
 
           my_db = MongoClient().test
           fs = GridFSBucket(my_db)
-          grid_in = fs.open_upload_stream_with_id(
+          with fs.open_upload_stream_with_id(
                 ObjectId(),
                 "test_file",
                 chunk_size_bytes=4,
-                metadata={"contentType": "text/plain"})
-          grid_in.write("data I want to store!")
-          grid_in.close()  # uploaded on close
+                metadata={"contentType": "text/plain"}) as grid_in:
+              grid_in.write("data I want to store!")
+          # uploaded on close
 
         Returns an instance of :class:`~gridfs.grid_file.GridIn`.
 
@@ -571,17 +620,27 @@ class GridFSBucket(object):
         """
         validate_string("filename", filename)
 
-        opts = {"_id": file_id,
-                "filename": filename,
-                "chunk_size": (chunk_size_bytes if chunk_size_bytes
-                               is not None else self._chunk_size_bytes)}
+        opts = {
+            "_id": file_id,
+            "filename": filename,
+            "chunk_size": (
+                chunk_size_bytes if chunk_size_bytes is not None else self._chunk_size_bytes
+            ),
+        }
         if metadata is not None:
             opts["metadata"] = metadata
 
         return GridIn(self._collection, session=session, **opts)
 
-    def upload_from_stream(self, filename, source, chunk_size_bytes=None,
-                           metadata=None, session=None):
+    @_csot.apply
+    def upload_from_stream(
+        self,
+        filename: str,
+        source: Any,
+        chunk_size_bytes: Optional[int] = None,
+        metadata: Optional[Mapping[str, Any]] = None,
+        session: Optional[ClientSession] = None,
+    ) -> ObjectId:
         """Uploads a user file to a GridFS bucket.
 
         Reads the contents of the user file from `source` and uploads
@@ -617,15 +676,21 @@ class GridFSBucket(object):
         .. versionchanged:: 3.6
            Added ``session`` parameter.
         """
-        with self.open_upload_stream(
-                filename, chunk_size_bytes, metadata, session=session) as gin:
+        with self.open_upload_stream(filename, chunk_size_bytes, metadata, session=session) as gin:
             gin.write(source)
 
-        return gin._id
+        return cast(ObjectId, gin._id)
 
-    def upload_from_stream_with_id(self, file_id, filename, source,
-                                   chunk_size_bytes=None, metadata=None,
-                                   session=None):
+    @_csot.apply
+    def upload_from_stream_with_id(
+        self,
+        file_id: Any,
+        filename: str,
+        source: Any,
+        chunk_size_bytes: Optional[int] = None,
+        metadata: Optional[Mapping[str, Any]] = None,
+        session: Optional[ClientSession] = None,
+    ) -> None:
         """Uploads a user file to a GridFS bucket with a custom file id.
 
         Reads the contents of the user file from `source` and uploads
@@ -663,11 +728,13 @@ class GridFSBucket(object):
            Added ``session`` parameter.
         """
         with self.open_upload_stream_with_id(
-                file_id, filename, chunk_size_bytes, metadata,
-                session=session) as gin:
+            file_id, filename, chunk_size_bytes, metadata, session=session
+        ) as gin:
             gin.write(source)
 
-    def open_download_stream(self, file_id, session=None):
+    def open_download_stream(
+        self, file_id: Any, session: Optional[ClientSession] = None
+    ) -> GridOut:
         """Opens a Stream from which the application can read the contents of
         the stored file specified by file_id.
 
@@ -698,7 +765,10 @@ class GridFSBucket(object):
         gout._ensure_file()
         return gout
 
-    def download_to_stream(self, file_id, destination, session=None):
+    @_csot.apply
+    def download_to_stream(
+        self, file_id: Any, destination: Any, session: Optional[ClientSession] = None
+    ) -> None:
         """Downloads the contents of the stored file specified by file_id and
         writes the contents to `destination`.
 
@@ -729,7 +799,8 @@ class GridFSBucket(object):
             for chunk in gout:
                 destination.write(chunk)
 
-    def delete(self, file_id, session=None):
+    @_csot.apply
+    def delete(self, file_id: Any, session: Optional[ClientSession] = None) -> None:
         """Given an file_id, delete this stored file's files collection document
         and associated chunks from a GridFS bucket.
 
@@ -755,10 +826,9 @@ class GridFSBucket(object):
         res = self._files.delete_one({"_id": file_id}, session=session)
         self._chunks.delete_many({"files_id": file_id}, session=session)
         if not res.deleted_count:
-            raise NoFile(
-                "no file could be deleted because none matched %s" % file_id)
+            raise NoFile("no file could be deleted because none matched %s" % file_id)
 
-    def find(self, *args, **kwargs):
+    def find(self, *args: Any, **kwargs: Any) -> GridOutCursor:
         """Find and return the files collection documents that match ``filter``
 
         Returns a cursor that iterates across files matching
@@ -806,7 +876,9 @@ class GridFSBucket(object):
         """
         return GridOutCursor(self._collection, *args, **kwargs)
 
-    def open_download_stream_by_name(self, filename, revision=-1, session=None):
+    def open_download_stream_by_name(
+        self, filename: str, revision: int = -1, session: Optional[ClientSession] = None
+    ) -> GridOut:
         """Opens a Stream from which the application can read the contents of
         `filename` and optional `revision`.
 
@@ -855,14 +927,18 @@ class GridFSBucket(object):
             cursor.limit(-1).skip(revision).sort("uploadDate", ASCENDING)
         try:
             grid_file = next(cursor)
-            return GridOut(
-                self._collection, file_document=grid_file, session=session)
+            return GridOut(self._collection, file_document=grid_file, session=session)
         except StopIteration:
-            raise NoFile(
-                "no version %d for filename %r" % (revision, filename))
+            raise NoFile("no version %d for filename %r" % (revision, filename))
 
-    def download_to_stream_by_name(self, filename, destination, revision=-1,
-                                   session=None):
+    @_csot.apply
+    def download_to_stream_by_name(
+        self,
+        filename: str,
+        destination: Any,
+        revision: int = -1,
+        session: Optional[ClientSession] = None,
+    ) -> None:
         """Write the contents of `filename` (with optional `revision`) to
         `destination`.
 
@@ -900,12 +976,13 @@ class GridFSBucket(object):
         .. versionchanged:: 3.6
            Added ``session`` parameter.
         """
-        with self.open_download_stream_by_name(
-                filename, revision, session=session) as gout:
+        with self.open_download_stream_by_name(filename, revision, session=session) as gout:
             for chunk in gout:
                 destination.write(chunk)
 
-    def rename(self, file_id, new_filename, session=None):
+    def rename(
+        self, file_id: Any, new_filename: str, session: Optional[ClientSession] = None
+    ) -> None:
         """Renames the stored file with the specified file_id.
 
         For example::
@@ -928,9 +1005,11 @@ class GridFSBucket(object):
            Added ``session`` parameter.
         """
         _disallow_transactions(session)
-        result = self._files.update_one({"_id": file_id},
-                                        {"$set": {"filename": new_filename}},
-                                        session=session)
+        result = self._files.update_one(
+            {"_id": file_id}, {"$set": {"filename": new_filename}}, session=session
+        )
         if not result.matched_count:
-            raise NoFile("no files could be renamed %r because none "
-                         "matched file_id %i" % (new_filename, file_id))
+            raise NoFile(
+                "no files could be renamed %r because none "
+                "matched file_id %i" % (new_filename, file_id)
+            )
